<?php
// $Id$

define('DITA_PROJECT_PREFIX_LENGTH', 6);
define('DITA_TOPIC_PREFIX', '');
define('DITA_TOPIC_POSTFIX', '.dita');
define('DITAMAP_DOCTYPE_STRING', '<!DOCTYPE map PUBLIC "-//OASIS//DTD DITA Map//EN"
 "http://docs.oasis-open.org/dita/v1.2/os/dtd1.2/technicalContent/dtd/map.dtd">');
define('DITAMAP_DOCTYPE', 'map');
define('DITAMAP_EXTERNAL_SUBSET_PUBLIC_IDENTIFIER',
  '-//OASIS//DTD DITA Map//EN');
define('DITAMAP_EXTERNAL_SUBSET_SYSTEM_IDENTIFIER',
  'http://docs.oasis-open.org/dita/v1.2/os/dtd1.2/technicalContent/dtd/map.dtd');

/**
 * Converts a freemind map to ditamap
 *
 * @param string $freemind_map an xml string of a afreemind map
 * (tested only with graphmind's output)
 * @param int $vid vocabulary id
 * @return FALSE on empty input xml. Valid ditamap else (without dtd reference!)
 */
function dita_freemind_to_ditamap_convert($freemind_map, $vid_array = array()) {
  $vid_array = make_array($vid_array);
  $error_message = 'Something went wrong during ditamap conversation';
  $xslt_url = drupal_get_path('module', 'dita') . '/freemind2ditamap.xsl';
  //$xml_object = new DOMDocument();
  if (!$freemind_map) {
    return FALSE;
  }

  $xml_string = dita_apply_xslt($freemind_map, file_get_contents($xslt_url));

  if (!$xml_string) {
    return $error_message;
  }

  $nids = dita_get_references($freemind_map);

  foreach ($nids as $nid) {
    $node = node_load($nid);
    $xml_string = str_replace('{dita_mindmap_title_placeholder-' . $nid . '}', dita_str_to_filename($node->title) . '_' . $nid, $xml_string);
  }

  $xml_object = dita_create_xml_object(
          $xml_string,
          DITAMAP_DOCTYPE,
          DITAMAP_EXTERNAL_SUBSET_PUBLIC_IDENTIFIER,
          DITAMAP_EXTERNAL_SUBSET_SYSTEM_IDENTIFIER);
  if (!$xml_object) {
    return $error_message;
  }
  $map_object = $xml_object->getElementsByTagName('map')->item(0);
  foreach ($vid_array as $vid) {
    // if there is no vocabulary, no reltable needed
    if ($vid != 0) {
      $reltable = dita_generate_reltable(dita_get_references($freemind_map), $vid);
      //an empty reltable(when no related vocabulary terms) can cause unecessary
      //warnings
      if ($reltable) {
        $reltable_object = new DomDocument();
        $reltable_object->loadXML('<?xml version="1.0" encoding="utf-8"?>' . $reltable);
        $reltable_root = $reltable_object->documentElement;
        $imported_reltable = $xml_object->importNode($reltable_root, TRUE);
        $map_object->insertBefore($imported_reltable);
      }
    }
  }

  return ($xml_object->saveXML());
}

/**
 * Applies an xslt schema to an xml file
 *
 * @param string $xml_string the xml
 * @param string $xslt_string the xslt
 * @param string working_dir Change working dir for the transformation. If false
 *               no directory change applies
 * @return string the output after conversion
 */
function dita_apply_xslt($xml_string, $xslt_string, $external_dtd = FALSE, $working_dir = FALSE) {

  if (!$xml_string) {
    return FALSE;
  }

  if (!$xslt_string) {
    drupal_set_message('No xslt given in dita_apply_xslt', 'warning');
    return $xml_string;
  }

  if ($working_dir) {
    $orig_dir = getcwd();
    chdir($working_dir);
  }

  $xml_object = new DOMDocument();
  if ($external_dtd) {
    $xml_object->loadXML($xml_string, LIBXML_DTDATTR);
  }
  else {
    $xml_object->loadXML($xml_string);
  }
  if (!$xml_object) {
    if ($working_dir) {
      chdir($orig_dir);
    }
    return FALSE;
  }

  $xslt_object = new DOMDocument();
  $xslt_object->loadXML($xslt_string);
  if (!$xslt_object) {
    if ($working_dir) {
      chdir($orig_dir);
    }
    return FALSE;
  }

  $xslt_processor = new XSLTProcessor;
  $xslt_processor->importStylesheet($xslt_object);
  if (!$xslt_processor) {
    if ($working_dir) {
      chdir($orig_dir);
    }
    return FALSE;
  }

  if ($working_dir) {
    chdir($orig_dir);
  }
  return $xslt_processor->transformToXML($xml_object);
}

/**
 * Creates a domdocument object from an xml string, with the desired
 * doctpye declaration
 *
 * @param string $xml input xml string
 * @param string $doctype the qualified name of the document type to create
 * @param string $public_id the external subset public identifier
 * @param string $system_id the external subset system identifier
 * @return object DomDocument object
 */

function dita_create_xml_object($xml, $doctype, $public_id, $system_id) {
  //insert doctype
  $implementation = new DOMImplementation();

  $dtd = $implementation->createDocumentType($doctype,
        $public_id,
        $system_id);

  $xml_object = $implementation->createDocument('', '', $dtd);

  $map_original = new DomDocument();
  $map_original->loadXML($xml);

  $child = $xml_object->importNode($map_original->documentElement, TRUE);
  $xml_object->appendChild($child);
  return $xml_object;
}

function make_array($variable) {
  if (is_array($variable)) {
    return $variable;
  }
  else {
    return array($variable);
  }
}


/**
 * Do the indentation of an XML file (if tidy extension is loaded)
 *
 * @param string
 *   Broken or not complete, wrongly indented xml
 * @return string
 *   The formatted XML
 */
function dita_format_xml_string($xml) {
  if (extension_loaded('tidy')) {
    $config = array(
      'indent' => TRUE,
      'output-xml' => TRUE,
      'input-xml' => TRUE,
      'wrap' => '1000'
    );

    $tidy = new tidy();
    $tidy->parseString($xml, $config, 'utf8');
    $tidy->cleanRepair();
    $xml = tidy_get_output($tidy);
  }

  return $xml;
}

/**
 * Creates a a string wich can be a machine-readable filename from a random string.
 * Practically that means it removes any non-alphabetical characters, and replaces
 * whitespaces to '_'
 *
 * @param string $str input string
 * @return string simplified string
 */
function dita_str_to_filename($str) {
  $return = '';
  if (empty($str)) {
    return '';
  }
  $str = str_split($str);
  foreach ($str as $key => $char) {
    $ord_char = ord($char);
    if ((dita_check_interval($ord_char, 48, 57))||
       (dita_check_interval($ord_char, 65, 90))||
       (dita_check_interval($ord_char, 97, 122))) {
      $return .= $char;
    }
    //convert whitespaces to -
    elseif ($char == ' ') {
      $return .= '-';
    }
  }
  return $return;
}


/**
 * Checks if a given number is in the given closed interval
 *
 * @param int $num The number
 * @param int $bottom lower limit point of interval
 * @param int $top upper limit point of the interval
 * @return bool if the point is in the interval
 */
function dita_check_interval($num, $bottom, $top) {
  if (($num >= $bottom) && ($num <= $top)) {
    return TRUE;
  }
  else {
    return FALSE;
  }
}


/**
 * Collects the node references from a given graphmind mindmap
 *
 * @param string $mindmap_xml A graphmind XML
 * @return array List of the node ids. FALSE if there were no node references
 */
function dita_get_references($mindmap_xml) {
  $xml_object = new DOMDocument();
  $xml_object->loadXML($mindmap_xml);
  $attributes = $xml_object->getElementsByTagName('attribute');

  $nodes = array();
  foreach ($attributes as $attribute) {
    if ($attribute->getAttribute('NAME') == 'nid') {
      $nodes[] = $attribute->getAttribute('VALUE');
    }
  }

  if (!empty($nodes)) {
    return $nodes;
  }

  return array();
}


/**
 *
 * @param array $reltable 3D array for the reltable
 * @param array $relheader list of rlospecs (content types)
 * @param bool $indent=FALSE if set, the output will be regurally indented
 *
 * @return string a formatted ditamap xml string
 */
function dita_render_reltable($reltable, $relheader, $indent = FALSE) {
  //@TODO check if this could be rewritten to use DOMDocument
  $reltable_str = "<reltable>\n";
  //relhead
  $reltable_str .= "<relheader>\n";
  if ((!is_array($relheader)) || (!is_array($reltable))) {
    return FALSE;
  }
  if (empty($relheader) || empty($reltable)) {
      //if no valid reltable can build, we don't build an empty reltable
      return '';
  }
  foreach ($relheader as $rolspec) {
    $reltable_str .= '<relcolspec type="'. $rolspec ."\"/>\n";
  }
  $reltable_str .= "</relheader>\n";
  //rows
  foreach ($reltable as $taxonomy_term => $relrow) {
    $reltable_str .= "<relrow>\n";
    //cells
    foreach ($relheader as $rolspec) {
      $reltable_str .= "<relcell>\n";
      if (isset($reltable[$taxonomy_term][$rolspec])) {
        $relcell = $reltable[$taxonomy_term][$rolspec];
        foreach ($relcell as $topicref) {
          $reltable_str .= '<topicref href="'. $topicref ."\"/>\n";
        }
      }
      $reltable_str .= "</relcell>\n";
    }
    $reltable_str .= "</relrow>\n";
  }
  $reltable_str .= "</reltable>\n";

  if ($indent) {
    return dita_format_xml_string($reltable_str);
  }
  return $reltable_str;
}

/**
 * Generates a reltable for a ditamap
 *
 * @param array $nodelist list of node objects
 * @param int $vid vocabulary id
 * @return string the reltable xml
 */
function dita_generate_reltable($nodelist, $vid) {
  $reltable = array();
  $rolspec_types = array();
  foreach ($nodelist as $nid) {
    $node = node_load($nid);
    $rolspec_type = dita_get_doctype($node);
    $terms = dita_get_taxonomy_term($node, $vid);
    //if the node has no terms, we don't do anything with it
     foreach ($terms as $term) {
      //reltable
      $reltable[$term->name][$rolspec_type][] = dita_str_to_filename($node->title) . '_' . $nid . DITA_TOPIC_POSTFIX;
      //collect rolspec types for the relheader
      if (!in_array($rolspec_type, $rolspec_types)) {
        $rolspec_types[] = $rolspec_type;
      }
    }
  }
  return dita_render_reltable($reltable, $rolspec_types);
}


/**
 * Determines the doctype of an xml document
 *
 * @param object $node a node object wich contains a valid xml in its body
 * @return string doctype
 */
function dita_get_doctype($node) {
  //@TODO further needs testing
  $xml_object = new DomDocument();
  $xml_object->loadXML(is_object($node) ? $node->body : $node);
  $dtype = $xml_object->doctype;
  return $dtype->nodeName;
}


/**
 * @todo Document
 */
function dita_get_taxonomy_term($node, $vid) {
  //@TODO Now we assume that the node has only one term... we have to create the
  //the vocabulary with install, and get the term of that vocabulary(?)
  $terms = taxonomy_node_get_terms_by_vocabulary($node, $vid);
  return $terms;
}

/**
 * Export functor for the DITA projects.
 */
class DITAProjectExporter {

  /**
   * The node object of the ditamap.
   *
   * @var stdClass
   */
  protected $node;

  /**
   * The base directory for the conversion.
   *
   * @var string
   */
  protected $basedir;
  protected $orig_basedir;

  /**
   * Prefix of the last conversion.
   *
   * @var string
   */
  public $prefix = NULL;

  public function getBasedir() {
    return $this->basedir;
  }

  public function setBasedir($basedir) {
    $this->basedir = $this->orig_basedir = $basedir;
  }

  public function resetBasedir() {
    $this->basedir = $this->orig_basedir;
  }

  /**
   * @constructor
   *
   * @param stdClass $node
   */
  public function __construct($node) {
    $this->node = $node;
  }

  /**
   * Invokes the functor.
   *
   * This function generates the
   *
   * @param array $vid_array list of vocabulary IDs.
   * 0 means that the reltable generation is disabled.
   * if int is passed, then that will be the id of the only related vocabulary
   * @return string Path for the .ditamap file.
   */
  public function __invoke($vid_array = 0) {

    $vid_array = make_array($vid_array);
    $this->resetBasedir();

    if (substr($this->basedir, -1) != '/') {
      $this->basedir .= '/';
    }
    $map_content = dita_freemind_to_ditamap_convert($this->node->body, $vid_array);
    $this->prefix = user_password(DITA_PROJECT_PREFIX_LENGTH);
    $map_filename = $this->prefix . $this->node->nid . '.ditamap';

    $this->basedir .= $this->prefix . '/';

    mkdir($this->basedir);

    $topiclist = dita_get_references($this->node->body);
    if (is_array($topiclist)) {
      foreach ($topiclist as $nid) {
        $topic_node = node_load($nid);

        /*@TODO REFACTOR: do the same with hooks*/
        //processing Poor Man's DITA's content types
        $body = $topic_node->body;

        watchdog('dita_debug', 'First checkpoint', $topic_node);

        if (module_exists('simple_dita_forms')) {

          if ($topic_node->type == 'dita_concept') {
            $dita_object = new SimpleDitaDocumentConcept($topic_node);
            $body = $dita_object-> getXmlString();
          }
          elseif ($topic_node->type == 'dita_task') {
            $dita_object = new SimpleDitaDocumentTask($topic_node);
            watchdog('dita_debug', 'Taskcheckpoint', $dita_objects);
            $body = $dita_object-> getXmlString();
          }
          elseif ($topic_node->type == 'dita_reference') {
            $dita_object = new SimpleDitaDocumentReference($topic_node);
            $body = $dita_object-> getXmlString();
          }
        }

        watchdog('dita_debug', 'Last checkpoint', $body);



        $topic_content = $body;
        $topic_filename = $this->basedir . dita_str_to_filename($topic_node->title) . '_' . $nid . DITA_TOPIC_POSTFIX;
        file_put_contents($topic_filename, $topic_content);
      }
    }

    //replace placeholders in the map with dynamic data
    $map_content = str_replace('{drupal_path_prefix}', $this->basedir, $map_content);

    file_put_contents($this->basedir . $map_filename, $map_content);
    return $this->basedir . $map_filename;
  }
}

//@TODO Check if the reltable generation is working properly
//@TODO Check if the OpenToolkit converting works correctly while there were
//      multiple changes in the generation method
//@TODO I get quite a few of these
//warning: XSLTProcessor::transformToXml() [xsltprocessor.transformtoxml]: I/O warning : failed to load external entity "/var/www/dita/sites/default/modules/dita/resource/messages.xml" in /var/www/dita/sites/default/modules/dita/dita.export.inc on line 105.
//at http://dita.pronovix.net/node/33
